# include <assert.h>
# include <stdio.h>
# include "gcc-externs.h"
# include "regset.ic"
# include "sdg.h"

regset regsInComp;

static void 
MarkPseudoRegsInExpr(rtx expr)
{
/* Identify all pseudo regs in the expr. and
 * also mark all operands in the expr as being in comp */

	RTX_CODE  code;
	int       regNum;

	expr->inComp = 1;
	switch (code = GET_CODE(expr)) {
			/* Since certain constant rtx exprs. are shared,
			 * cannot mark this rtx.  Have to use some other
			 * technique to identify this operand as being in Comp */
	case CONST_INT :
/*
		expr->inComp = 0;
*/
		break;

			/* The following cases are constant expressions
			 * which do not use any pseudo regs. */
	case CONST_DOUBLE :
	case CONST_STRING :
	case SYMBOL_REF :
	case LABEL_REF :
	case CONST :
	case HIGH :
	case CC0 :
	case PC :
	case SCRATCH :
		break;

	case MEM :
			/* Not supposed to come here */
		assert(0);
		break;

	case REG :
			/* This corresponds to the reg 'expr' */
		regNum = REGNO(expr);
		if (regNum > LAST_VIRTUAL_REGISTER)
			SetBit(regsInComp, regNum);
		break;

	case SUBREG :
		regNum = REGNO(SUBREG_REG(expr));
		if (regNum > LAST_VIRTUAL_REGISTER) {
			SetBit(regsInComp, regNum);
			(SUBREG_REG(expr))->inComp = 1;
		}
		break;

	case STRICT_LOW_PART :
		regNum = REGNO(SUBREG_REG(XEXP(expr, 0)));
		if (regNum > LAST_VIRTUAL_REGISTER) {
			SetBit(regsInComp, regNum);
			(XEXP(expr, 0))->inComp = 1;
			(SUBREG_REG(XEXP(expr, 0)))->inComp = 1;
		}
		break;

			/* Binary arithmetic operations */
	case PLUS :
	case LO_SUM :
	case MINUS :
	case COMPARE :
	case MULT :
	case DIV : 
	case UDIV :
	case MOD : 
	case UMOD :
	case SMIN :
	case SMAX :
	case UMIN :
	case UMAX :
			/* Binary bit operations */
	case AND :
	case IOR :
	case XOR :
	case ASHIFT :
	case LSHIFTRT :
	case ASHIFTRT :
	case ROTATE :
	case ROTATERT :
			/* Binary comparison operations */
	case EQ :
	case NE :
	case GT :
	case GTU :
	case LT :
	case LTU :
	case GE :
	case GEU :
	case LE :
	case LEU :
		MarkPseudoRegsInExpr(XEXP(expr, 0));
		MarkPseudoRegsInExpr(XEXP(expr, 1));
		break;

			/* Unary arithmetic operations */
	case NOT :
	case NEG :
	case ABS :
	case SQRT :
	case FFS :
		MarkPseudoRegsInExpr(XEXP(expr, 0));
		break;

			/* Cond jump. */
	case IF_THEN_ELSE :
		MarkPseudoRegsInExpr(XEXP(expr, 0));	/* condition */
		MarkPseudoRegsInExpr(XEXP(expr, 1));	/* then-clause */
		MarkPseudoRegsInExpr(XEXP(expr, 2));	/* else-clause */
		break;

	case COND :
		fprintf(stderr, 
				"COND is not supposed to appear as an instr. pattern\n");
		abort();
		break;

			/* Bit-field operations */
			/* The bits are extracted from XEXP(expr, 0) */
	case SIGN_EXTRACT :
	case ZERO_EXTRACT :
		MarkPseudoRegsInExpr(XEXP(expr, 0)); 
		break;

			/* Conversions */
	case FIX :
	case UNSIGNED_FIX :
	case FLOAT_EXTEND :
	case FLOAT_TRUNCATE :
			/* This is floating point code and will anyway
			 * be assigned fp regs.  Dont bother */
		break;

			/* Call.  Cannot come here. */
	case CALL :
		assert(0);
		break;

			/* Conversions */
	case SIGN_EXTEND :
	case ZERO_EXTEND :
		MarkPseudoRegsInExpr(XEXP(expr, 0)); 
		break;

			/* Conversions */
	case FLOAT :
	case UNSIGNED_FLOAT :
	case TRUNCATE :
		MarkPseudoRegsInExpr(XEXP(expr, 0)); 
		break;

			/* Unhandled codes */
	default :
		fprintf(stderr, "Unhandled case in \'MarkPseudoRegsInExpr\'\n");
		abort();
	}
}

static void
MarkPseudoRegsIn_SET_rtx(rtx insn)
{
/* Call instructions always execute in the ld-st subsystem */

	if (GET_CODE(SET_SRC(insn)) != CALL) {
		MarkPseudoRegsInExpr(SET_SRC(insn));
		MarkPseudoRegsInExpr(SET_DEST(insn));
	}
	else
		insn->inComp = 0;
}

static void
MarkPseudoRegsInInstr(rtx insn)
{
/* Identify all registers used/set in this instruction */

	rtx   body, dest;

	switch (GET_CODE(insn)) {
	case ADDR_VEC :
	case ADDR_DIFF_VEC :
	case RETURN :
	case USE :
	case CLOBBER :
	case TRAP_IF : /* Not yet handled */
	case ASM_INPUT :
	case UNSPEC :
	case UNSPEC_VOLATILE :
		break;

	case SET :
		MarkPseudoRegsIn_SET_rtx(insn);
		break;

	case CALL :
	case PARALLEL :
	case SEQUENCE :
		assert(0);
		break;

	default :
		fprintf(stderr, "Unhandled case in \'MarkPseudoRegsInInstr\'\n");
		abort();
	}
}

int
OnlySuppliedByLoads(SdgNode *storeVal)
{
/* Need to do this because of 'gcc'.  When this store is only supplied by
 * loads and the load only supply this store, in certain cases, 'gcc' inserts
 * copies if I load into a fp reg and store the fp reg.  So, just push such
 * store values into LdSt.  We dont lose anything due to this */

	Iterator iter;
	SdgNode *src;

	for (src = (SdgNode *)InitIter(iter, storeVal->srcs);
		 src != NULL;
		 src = (SdgNode *)GetNext(iter))
	{
			/* Make sure 'src' is a load */
		if (!src->isLoadOp)
			return 0;
	}

	for (src = (SdgNode *)InitIter(iter, storeVal->srcs);
		 src != NULL;
		 src = (SdgNode *)GetNext(iter))
	{
			/* Make sure 'src' only supplies 'storeVal' */
		if (GetHead(src->dsts) != GetTail(src->dsts))
			return 0;
	}

	storeVal->inLdSt = 1;
	for (src = (SdgNode *)InitIter(iter, storeVal->srcs);
		 src != NULL;
		 src = (SdgNode *)GetNext(iter))
	{
		src->inLdSt = 1;
		src->instr->inComp = 0;
	}

	return 1;
}

void 
MarkCompInstructions()
{
/* Mark instrs. not in the 'ld-st' subsystem to be in 
 * the 'Comp' subsystem.
 * Also, identify regs. used by these instrs.
 * They would have to be allocated fp regs */

	SdgNode  *node = sdg.head;
	Iterator  iter;
	int       i;

	AllocRegset(regsInComp);
	InitRegset(regsInComp);

	while (node != NULL) {
				/* For store values, need not mark anything.
				 * The source reg. will be marked by other guys.
				 * The dest expression SHOULDN'T be marked since
				 * it is part of the address computation.
				 */
		if (node->isStoreVal && !node->inLdSt) {
			if (!OnlySuppliedByLoads(node))
				node->instr->inComp = 1;
		}
				/* If node is marked 'conv', then it is the fp
				 * part of the conversion operation and it will
				 * always execute in 'Comp'.  However, whether the
				 * integer part of the conversion executes in 'LdSt' or
				 * 'Comp' will be decided by the integer part itself */
		else if (!node->inLdSt && !node->isAConvNode) {
				/* Mark this instr as being in the 'comp' subsystem */
			node->instr->inComp = 1;

			if (node->isLoadOp) {
				node = node->next;
				continue;
			}
				/* Need to identify all regs in 'node' 
				 * as they should now be allocated fp registers. */
			switch (GET_CODE(node->instr)) {
				case INSN :
					MarkPseudoRegsInInstr(PATTERN(node->instr));
					break;

				case SET :
					MarkPseudoRegsIn_SET_rtx(node->instr);
					break;

				case JUMP_INSN :
						/* Indirect jumps get handled by this */
					MarkPseudoRegsInInstr(PATTERN(node->instr));
					break;

				case CALL_INSN :
						/* Calls.  All call instructions
						 * execute in the ld-st subsystem
						 * Later, can allow some of these to execute
						 * in the comp subsystem */
					node->instr->inComp = 0;
					break;

				default :
					break;
			}
		}

		node = node->next;
	}

	if (dump_comp_regs) {
		fputc('\n', sdg_dump_file);
		for (i = 0; i < max_regno; i++)
			if (BitIsSet(regsInComp, i)) {
				fprintf(sdg_dump_file, "Reg %d is a fp reg.", i);
				if (regno_reg_rtx[i] == NULL)
					fprintf(sdg_dump_file, " -- NULL rtx\n");
				else if (INTEGRAL_MODE_P(GET_MODE(regno_reg_rtx[i])))
					fprintf(sdg_dump_file, " -- INT mode\n");
				else
					fprintf(sdg_dump_file, " -- FP mode\n");
			}
	}
}
