# include <string.h>
# include <ctype.h>
# include <assert.h>
# include "gcc-externs.h"

extern FILE *asm_out_file;

static char *fileName = NULL;

void
SetFileName()
{
/* Filter dump_base_name of all special characters */

	int i = 0, j = 0;

	if (fileName != NULL)
		free(fileName);

	fileName = (char *)malloc(strlen(dump_base_name) + 1);
	while (dump_base_name[i] != '\0') {
		if (isalnum(dump_base_name[i]) || (dump_base_name[i] == '_'))
			fileName[j++] = dump_base_name[i];
		i++;
	}

	fileName[j] = '\0';
}

void
AssembleCounters()
{
	static int count = 0;
	static int done = 0;

	SetFileName();
    fprintf(asm_out_file, "\t.data\n");
    fprintf(asm_out_file, "\t.align 2\n");
    fprintf(asm_out_file, "cnt_%s_%s:\n", currentFnName, fileName);
    fprintf(asm_out_file, "\t.word 0\n");
	if (n_basic_blocks > 1)
		fprintf(asm_out_file, "\t.space %d\n", (n_basic_blocks - 1)*4);
	fprintf(asm_out_file, "\t.sdata\n");
	fprintf(asm_out_file, "\t.align 2\n");
	fprintf(asm_out_file, "SSS_%s:\n", currentFnName);
	fprintf(asm_out_file, "\t.ascii \"%s\\000\"\n", currentFnName);
	if (!done) {
		fprintf(asm_out_file, "\t.sdata\n");
		fprintf(asm_out_file, "\t.align 2\n");
		fprintf(asm_out_file, "SSS_%s:\n", fileName);
		fprintf(asm_out_file, "\t.ascii \"%s\\000\"\n", dump_base_name);
		fprintf(asm_out_file, "\n\t.extern SSS_numCounters, 2\n");
		fprintf(asm_out_file, "\t.extern SSS_counters, 4\n");
		fprintf(asm_out_file, "\t.extern SSS_fileName, 4\n");
		fprintf(asm_out_file, "\t.extern SSS_fnName, 4\n\n");
		done = 1;
	}
	count++;

		/* Force $31 to be saved */
	regs_ever_live[GP_REG_FIRST+31] = 1;
}

void
InitProfilerForFunction()
{
    fprintf(asm_out_file, "\tla \t$9, cnt_%s_%s\n", currentFnName, fileName);
    fprintf(asm_out_file, "\tlw \t$8, 0($9)\n");
    fprintf(asm_out_file, "\tbne\t$8, $0, SSS_LL_%s\n", currentFnName);
	fprintf(asm_out_file, "\tli \t$8, %d\n", n_basic_blocks);
    fprintf(asm_out_file, "\tsh \t$8, SSS_numCounters\n");
    fprintf(asm_out_file, "\tsw \t$9, SSS_counters\n");
	fprintf(asm_out_file, "\tla \t$8, SSS_%s\n", currentFnName);
	fprintf(asm_out_file, "\tsw \t$8, SSS_fnName\n");
	fprintf(asm_out_file, "\tla \t$8, SSS_%s\n", fileName);
	fprintf(asm_out_file, "\tsw \t$8, SSS_fileName\n");
    fprintf(asm_out_file, "\tjal\tRegisterCounters\n");
    fprintf(asm_out_file, "SSS_LL_%s:\n", currentFnName);
}

void
OutputCounterUpdate(int index)
{
    fprintf(asm_out_file, "\tla \t$9, cnt_%s_%s+%d\n", 
						  currentFnName, fileName, 4*index);
    fprintf(asm_out_file, "\tlw \t$8, 0($9)\n");
	fprintf(asm_out_file, "\tadd\t$8, $8, 0x1\n");
	fprintf(asm_out_file, "\tsw \t$8, 0($9)\n");
}

static int
InsnIsInBB(int bbId, rtx insn)
{
    rtx tmp = basic_block_head[bbId];
    rtx end = basic_block_end[bbId];

    while (1) {
		if (insn == tmp)
		   	return 1;

		if (tmp == end || tmp == NULL)
		   	return 0;

		tmp = NEXT_INSN(tmp);
    }
}

int
GetBasicBlockId(rtx insn, int possibleBBid)
{
    int i, last;
	extern int *uid_block_number;
	extern int  max_uid_for_flow;

	if (   INSN_UID(insn) < max_uid_for_flow
		&& uid_block_number[INSN_UID(insn)] > 0
		&& uid_block_number[INSN_UID(insn)] < n_basic_blocks)
	{
		return uid_block_number[INSN_UID(insn)];
	}

	last = possibleBBid + 3;
	if (last > n_basic_blocks)
		last = n_basic_blocks;

    for (i = possibleBBid; i < last; i++) {
		if (InsnIsInBB(i, insn))
			return i;
    }

    for (i = n_basic_blocks-1; i >= 0; i--) {
		if (InsnIsInBB(i, insn))
			return i;
    }

	if (GET_CODE(insn) != JUMP_INSN && NEXT_INSN(insn) != NULL)
		return GetBasicBlockId(NEXT_INSN(insn), possibleBBid);
	else if (PREV_INSN(insn) != NULL && GET_CODE(PREV_INSN(insn)) != JUMP_INSN)
		return GetBasicBlockId(PREV_INSN(insn), possibleBBid);
	else {
		fprintf(stderr, "Insn's bb not found!!\n");
		return possibleBBid;
	}
}
