# include <stdlib.h>
# include <assert.h>
# include "gcc-externs.h"
# include "list.h"
# include "sdg.h"
# include "bb.h"
# include "ud.h"
# include "regset.ic"

		/* This regset identifies all those regs. which have defined in 
		 * the active basic block.  Set in 'IdentifyDefnsInBB' and used 
		 * immediately in 'InitBasicBlock'.  No longer valid after this */
regset *regsDefinedInBB;

		/* The static dependency graph */
Graph   sdg;

		/* This regset identifies all those regs. which are
		 * to be allocated registers in the Comp. subsystem */
extern regset regsInComp;

		/* The list of instructions from which slicing begins */
extern List sliceNodeList;

static int nodeCount = 0;

# ifdef __GNUC__
inline
# endif
void
InitSDG()
{
	sdg.head = sdg.tail = NULL;
	nodeCount = 0;
# ifdef COPYOPT_SLICING
	{
		extern void InitDummyArgumentNodes();
		InitDummyArgumentNodes();
	}
# endif
}

SdgNode *
AllocNewSDGnode(rtx insn)
{
	SdgNode *node = (SdgNode *)malloc(sizeof(SdgNode));

	node->next = NULL;
	node->inLdSt = 0;
	node->isACall = 0;
	node->isMemOp = 0;
	node->isLoadOp = 0;
	node->isStoreOp = 0;
	node->isStoreVal = 0;
	node->isAddr = 0;
	node->isMoveOp = 0;
	node->isAConvNode = 0;
	node->srcUsesHardIntReg = 0;
	node->dstUsesHardIntReg = 0;
	node->srcUsesIntArgReg = 0;
	node->dstUsesIntArgReg = 0;
	node->srcUsesIntRetReg = 0;
	node->dstUsesIntRetReg = 0;
	node->visited = 0;
	node->inProfile = 0;
# ifdef COPYOPT_SLICING
	node->copiesAreNeeded = 0;
	node->copyInserted = 0;
	node->mightBeInLdst = 0;
	node->copiesMightBeNeeded = 0;
	node->shouldDuplicate = 0;
	node->insertedInList = 0;
	node->replacementReg = NULL;
	node->copyingCost = node->duplicatingCost = 0.0;
# endif
	node->inCompoundRtx = 0;
	node->compoundRtx = NULL;

	node->id = nodeCount++;
	node->instr = insn;
	node->nodeWeight = 1.0;
	Init(node->dsts);
	Init(node->srcs);

	return node;
}

# ifdef __GNUC__
inline
# endif
static int
IsAMoveInstr(rtx instr)
{
	rtx src, dst;

	switch (GET_CODE(instr)) {
		case SET :
			src = SET_SRC(instr);
			dst = SET_DEST(instr);
			return (GET_CODE(src) == REG && GET_CODE(dst) == REG);

		case INSN :
			if (GET_CODE(PATTERN(instr)) == SET) {
				src = SET_SRC(PATTERN(instr));
				dst = SET_DEST(PATTERN(instr));
				return (GET_CODE(src) == REG && GET_CODE(dst) == REG);
			}
			else
				return 0;

		default :
			return 0;
	}
}

SdgNode *
InsertInsnIntoSDG(rtx insn)
{
	SdgNode *node = AllocNewSDGnode(insn);
	if (sdg.head == NULL)
		sdg.head = sdg.tail = node;
	else {
		sdg.tail->next = node;
		sdg.tail = node;
	}
	node->isMoveOp = IsAMoveInstr(insn);
	return node;
}

void
Initialise()
{
/* Make an initial pass over all the basic blocks of the function and
 * build the control-flow graph of the function.
 * Also, identify all pseudo reg definitions */

	int         i;
	regset      tmp;
	extern int numDefns;

	extern void InitFreePseudoRegList();
	extern void ReadInBBexecutionCounts();
	extern int  max_reg_num();

		/* Since the # of basic blocks is known, all bbs
		 * can be allocated at the same time as an array */
	ALLOCATE_BASIC_BLOCKS;

	max_regno = max_reg_num();

		/* Initialise the static dependency graph */
	InitSDG();

		/* Initialise the list of free pseudo registers */
	InitFreePseudoRegList();

		/* Initialise the load-store list */
	Init(sliceNodeList);

		/* Initialise the definitions array */
	InitDefns();

		/* Initialise each basic block */
	for (i = 0; i < n_basic_blocks; i++)
		InitBasicBlock(i);

		/* Build the CFG.  Can't build until the bbs are all initialised */
	for (i = 0; i < n_basic_blocks; i++)
		SetControlFlowEdges(GetBasicBlock(i));

# ifndef PLAIN_SLICING
		/* Compute probabilities of each bb getting executed */
	ComputeBBexecutionProbabilities();

	ReadInBBexecutionCounts();
# endif

	regsDefinedInBB = (regset *)calloc(n_basic_blocks, sizeof(regset));
	tmp = (regset)malloc(n_basic_blocks * regset_bytes);

		/* Initialise the pred & succ lists in all basic blocks.
		 * Also, go through each basic block, and identify all defns */
	for (i = 0; i < n_basic_blocks; i++) {
		regsDefinedInBB[i] = tmp;
		tmp += regset_bytes/sizeof(*tmp);
		IdentifyDefnsInBB(i);
	}

		/* Only after all basic blocks have been traversed and
		 * defns identified is 'numDefns' available.. */
	COMPUTE_DEFN_SET_SIZE;

	for (i = 0; i < n_basic_blocks; i++)
		AllocMemory(GetBasicBlock(i));
}

void
ComputeReachingDefns()
{
/* Use a worklist algorithm to compute the ud chains(reaching defns.) */

	int  		i;
	BasicBlock *bb;
	List 	    workList;
	BBset       bbset;

	AllocBBset(bbset);
	InitBBset(bbset);

/* Initially, put all the basic blocks on the worklist */
	Init(workList);
	for (i = 0; i < n_basic_blocks; i++) {
		BB_SetBit(bbset, i);
		Append(&workList, GetBasicBlock(i));
	}

/* Now, remove elements one by one and proceed */
	while ((bb = RemoveHead(&workList)) != NULL) {
			/* Process 'bb'
			 *   (i) Firstly, compute for each active preg, compute
			 *        'defsIn' as a union of defsOut of 'bb's predecessors
			 *  (ii) Next, look at each instr. of the bb. to compute
			 *        'defsOut' 
			 * (iii) If 'defsOut' for any active preg has changed from
			 *        the last time, add 'bb's successors to the worklist
			 */
		BB_UnsetBit(bbset, bb->id);
		ComputeDefsIn(bb);
		if (ComputeDefsOut(bb))
			AddSuccessorsToWorkList(bb, &workList, bbset);
	}

	FreeBBset(bbset);
}

# if 0

void
PrintReachingDefns()
{
	int	i;
	extern void PrintDefns(void);

	PrintDefns();
	for (i = 0; i < n_basic_blocks; i++)
		PrintDefsIn(GetBasicBlock(i));
}

# endif

void
PrintCfg()
{
	int i;

	fprintf(sdg_dump_file, "\n\n########## NEW FUNCTION ##########\n\n");
	for (i = 0; i < n_basic_blocks; i++)
		PrintCfgEdges(GetBasicBlock(i));
}

# ifdef __GNUC__
inline
# endif
void
AddSdgEdge(SdgNode *defn, SdgNode *use)
{
/* Add 'defn' as a source of 'use'
 * Add 'use' as a destination of 'defn' */

/* Do not insert.  Some code depends on edges
 * being in the same order as they are added */

	Append(&defn->dsts, use);
	Append(&use->srcs, defn);
}

void
AddSdgEdges(Preg *reg, SdgNode *use)
{
	int   i;
	Defn *defn;
	int   flag = 0;

	for (i = 0; i < numDefns; i++)
		if (Defn_BitIsSet(reg->tmp, i)) {
			flag = 1;
			defn = GetDefn(i);
			AddSdgEdge(defn->node, use);
		}

# ifdef COPYOPT_SLICING
	if (flag == 0) {
		extern SdgNode *GetDummyArgumentNode(int regNum);
		SdgNode *dummy;

		dummy = GetDummyArgumentNode(reg->regNo);
		if (dummy != NULL)
			AddSdgEdge(dummy, use);
	}
# endif
}

void
AddDependencyEdges()
{
	int i;

	for (i = 0; i < n_basic_blocks; i++)
		IdentifyDependenciesInBB(GetBasicBlock(i));
}

void
PrintSDG()
{
	SdgNode  *tmp = sdg.head;
	Iterator  iter;
	SdgNode  *tmptmp;

	extern void print_single_rtl(FILE *, rtx);

	while (tmp != NULL) {
		fprintf(sdg_dump_file, "#####################\n");
		fprintf(sdg_dump_file, "## Node %d : ", tmp->id);
		if (tmp->inLdSt)
			fprintf(sdg_dump_file, "<< Ld-St >>");
		else
			fprintf(sdg_dump_file, "<< Comp  >>");

		if (!tmp->instr->inComp)
			fprintf(sdg_dump_file, "<< INSN -- Ld-St >>");
		else
			fprintf(sdg_dump_file, "<< INSN -- Comp  >>");

		print_single_rtl(sdg_dump_file, tmp->instr);

			/* Print instrs. which produce values that I consume */
		fprintf(sdg_dump_file, "\n## SOURCES : ");
		tmptmp = (SdgNode *)InitIter(iter, tmp->srcs);
		while (tmptmp != NULL) {
			fprintf(sdg_dump_file, "%d ", tmptmp->id);
			tmptmp = (SdgNode *)GetNext(iter);
		}

			/* Print instrs. that consume values that I produce */
		fprintf(sdg_dump_file, "\n## DESTS : ");
		tmptmp = (SdgNode *)InitIter(iter, tmp->dsts);
		while (tmptmp != NULL) {
			fprintf(sdg_dump_file, "%d ", tmptmp->id);
			tmptmp = (SdgNode *)GetNext(iter);
		}

		fputc('\n', sdg_dump_file);

		tmp = tmp->next;
	}
}

void
FreeSDG()
{
	SdgNode *next;
	SdgNode *tmp = sdg.head;

	while (tmp != NULL) {
		next = tmp->next;
		FreeList(&tmp->srcs);
		FreeList(&tmp->dsts);
		free(tmp);
		tmp = next;
	}
	sdg.head = sdg.tail = NULL;
# ifdef COPYOPT_SLICING
	{
		extern void FreeDummyArgumentNodes();
		FreeDummyArgumentNodes();
	}
# endif
}

void
CleanUp()
{
/* After this point, only thing that is needed is the
 * the static dependency graph.  Free all other
 * structures - basic blocks, pregs, defns., etc.
 */
	free(regsDefinedInBB[0]);
	free(regsDefinedInBB);
	regsDefinedInBB = 0;
}

void
BuildStaticDependencyGraph()
{
/*
 *   (i) Build the cfg
 *  (ii) Compute reaching defns
 * (iii) Identify dependencies and build the dependence graph
 *  (iv) Optimize the # of copies needed
 *   (v) Free memory
 */

	extern void OptimizeNumberOfCopies();

		/* Make the initial pass over the basic blocks 
		 * to build the cfg and identify all defns */
	Initialise();

		/* Print the cfg */
	if (dump_cfg)
		PrintCfg();

 	    /* Build use-def chains */
	ComputeReachingDefns();

# if 0
		/* Print all reaching defns */
	PrintReachingDefns();
# endif

		/* Now, make an extra pass over the basic blocks to
		 * add the dependency edges */
	AddDependencyEdges();

# ifdef ARGOPT_SLICING
		/* Minimize the number of copies that might be needed
		 * to copy values into/from argument/return registers */
	OptimizeNumberOfCopies();
# endif

		/* Clean up.. Free memory allocated for reaching defns. */
	CleanUp();
}

# ifdef YANOS_STATS

BasicBlock *
GetBasicBlockOfNode(SdgNode *node)
{
	rtx insn;
	int uid;

	extern int  max_uid_for_flow;
	extern int *uid_block_number;

	insn = (node->inCompoundRtx) ? node->compoundRtx : node->instr;
	uid = INSN_UID(insn);
	if (uid > max_uid_for_flow)
		return NULL;
	else {
		int bbId = uid_block_number[uid];
		return (bbId < n_basic_blocks) ? GetBasicBlock(bbId) 
									   : (BasicBlock *)NULL;
	}

	assert(0);
}	

void
FindSelfLoopsInSDG()
{
	SdgNode *node;

	for (node = sdg.head; node != NULL; node = node->next) {
		Iterator    iter;
		SdgNode    *src;
		int         selfLoopPresent;
		int         allOtherParentsOutside;
		int         numSources;
		BasicBlock *nodeBB;

			/* SKIP DUMMY ARG NODES */
		if (node->id < 4)
			continue;

		nodeBB = GetBasicBlockOfNode(node);
		if (nodeBB == NULL) {
			fprintf(stderr, "Cannot get info. for node %d\n", node->id);
			continue;
		}

		selfLoopPresent = 0;
		allOtherParentsOutside = 1;
		numSources = 0;
		for (src = InitIter(iter, node->srcs);
		     src != NULL;
			 src = GetNext(iter))
		{
				/* SKIP DUMMY ARG NODES */
			if (src->id < 4)
				continue;

			if (src == node)
				selfLoopPresent = 1;
			else {
				BasicBlock *srcBB = GetBasicBlockOfNode(src);

				numSources++;
				if (srcBB != NULL) {
					if (nodeBB->loopDepth > 0) {
						if (   (srcBB->loopDepth == nodeBB->loopDepth)
							&& (srcBB->innerLoop == nodeBB->innerLoop))
							allOtherParentsOutside = 0;
					}
				}
				else {
					fprintf(stderr, "Possible incorrect info. for node %d\n", 
							node->id);
				}
			}
		}

		if (selfLoopPresent) {
			fprintf(stderr, "###########################\n");
			fprintf(stderr, "Node %d has a self-loop.\n", node->id);
			if (nodeBB->loopDepth == 0) {
				fprintf(stderr, "It is not nested in a loop\n");
			}
			else {
				fprintf(stderr, "It is INSIDE a loop, has %d other sources\n",
						numSources);
				if (allOtherParentsOutside)
					fprintf(stderr, "All those are outside the loop\n");
				else
					fprintf(stderr, "Some are inside the same loop\n");
			}
			fprintf(stderr, "The instruction is :\n");
			print_single_rtl(stderr, node->instr);
			fprintf(stderr, "\n");
		}
	}
}
# endif

void 
SplitInstructionStream()
{
/*  (i) Traverse the whole function(list of basic blocks) 
 *       and build the static dependence graph.
 * (ii) Next, slice the graph into the 'ld-st' and 'comp'
 *       components
 */

	extern void SliceDependencyGraph();
	extern void CheckForRegsInBothSubsystems();
	extern void MarkCompInstructions();

	fprintf(stderr, "### Function %s ###\n", currentFnName);
		/* If the function is a trivial 'return'
		 * then don't need to do anything */
	if (n_basic_blocks == 0)
		return;

		/* Firstly, build the static dependency graph */
	BuildStaticDependencyGraph();

# ifdef YANOS_STATS
	FindSelfLoopsInSDG();
	if (dump_sdg)
		PrintSDG();
	return;
# else
		/* Now, slice the graph into the 'ld-st' and the 'comp' components */
	SliceDependencyGraph();

		/* Mark all instructions in the 'comp' subsystem
		 * Also, Identify registers in the 'comp' subsystem.
		 * These have to be allocated fp registers. */
	MarkCompInstructions();

		/* Check if there are pseudo regs which are used
		 * in both subsystems and introduce a new one for the
		 * use of the pseudo in the ld-st subsystem */
	CheckForRegsInBothSubsystems();

		/* Print the static dependency graph */
	if (dump_sdg)
		PrintSDG();

		/* Free the CFG */
	FreeCfg();

		/* Free the defns */
	FreeDefns();

		/* Free the static dependency graph too */
	FreeSDG();
# endif
}

void CleanUpTheMess()
{
	free(regsInComp);
	regsInComp = NULL;
}
